import pytest
import numpy as np
from sktime.forecasting.ets import AutoETS
from metats.forecasters.utils import is_sktime_forecaster, generate_sktime_prediction

def test_sktime_wrapper():
    y = 10 + np.random.normal(size=(2, 55, 5))
    forecaster = AutoETS(auto=True, sp=4, n_jobs=-1)
    assert is_sktime_forecaster(forecaster)
    predictions = generate_sktime_prediction(forecaster, y, fh=4, forecast_dim=0)
    assert predictions.shape == (2, 4)


import pytest
import numpy as np

from metats.forecasters.utils import is_darts_forecaster
from metats.forecasters.wrappers import DartsForecasterWrapper
from darts.models.forecasting.exponential_smoothing import ExponentialSmoothing as ETS
from darts.models import Theta, AutoARIMA, TBATS

def test_darts():
    ets_model = Theta()
    assert is_darts_forecaster(ets_model)
    wrapped_model = DartsForecasterWrapper(ets_model)
    data = np.random.uniform(size=(4, 55, 5))
    predictions = wrapped_model.predict(data, fh=7, forecast_dim=1)
    assert predictions.shape == (4, 7)


import sktime
import numpy as np
from sktime.base import BaseEstimator
from darts.models.forecasting.forecasting_model import ForecastingModel

def is_sktime_forecaster(forecaster):
    """
    Check if a given object is a sktime forecaster
    """
    return isinstance(forecaster, BaseEstimator)


def is_darts_forecaster(forecaster):
    """
    Check if a given object is a darts forecaster
    """
    return isinstance(forecaster, ForecastingModel)

def is_nixtla_stats_forecaster(forecaster):
    from statsforecast import StatsForecast
    """
    Check if a given object is a nixtla statistical forecaster
    """
    return isinstance(forecaster, StatsForecast)

    


import numpy as np
from metats.forecasters import BaseForecaster

# Naive Forecaster
class SeasonalNaiveForecaster(BaseForecaster):
    """
    Seasonal Naive Forecaster
    """
    def __init__(self):
        self.name = 'Seasonal Naive Forecaster'
    
    def predict(self, y, fh):
        """
        forecasting using the naive method
        inputs:
            y: the time series to be forecasted numpy array (num_series x series_length x covariates_dim)
            fh: the number of steps to forecast
        """
        if y.ndim < 3:
            z = y[:, :, np.newaxis]
        predictions = z[:, -fh:, :]
        return predictions


import numpy as np
from metats.forecasters.statistical import SeasonalNaiveForecaster

def test_forecaster():
    data = np.random.rand(10, 55, 5)
    forecaster = SeasonalNaiveForecaster()
    predictions = forecaster.predict(data, horizon=7)
    assert predictions.shape == (10, 7, 5)
    assert np.all(predictions[:, :, :] == data[:, -7:, :])


import torch
import torch.nn as nn
import warnings
warnings.filterwarnings('ignore')

class LSTMEncoder(nn.Module):
  """
  A general class for stacked lstm encoder
  """
  def __init__(self, input_size, latent_size, hidden_size, num_layers, directions):
    """
    inputs:
        input_size: dimension of input series
        latent_size: dimension of latent representation
        hidden_size: dimension of lstm hidden cells
        num_layers: number of stacked lstm
        directions: 1 for ordinary and 2 for bi-directional lstm
    """
    super().__init__()

    is_bidirectional = (directions == 2)

    self.lstm = nn.LSTM(input_size, hidden_size=hidden_size,
                        bidirectional=is_bidirectional, num_layers=num_layers,
                        proj_size=latent_size)
    
    self.latent_size = latent_size
    self.num_layers = num_layers
    self.directions = directions
    self.hidden_size = hidden_size
  
  def get_initial(self, batch_size):
    h_size = (self.directions * self.num_layers, batch_size, self.latent_size)
    c_size = (self.directions * self.num_layers, batch_size, self.hidden_size)
    h0 = torch.zeros(h_size)
    c0 = torch.zeros(c_size)
    return h0, c0
  
  def forward(self, Y):
    bsize = Y.size(0)
    Y = Y.permute(1, 0, 2)
    h0, c0 = self.get_initial(bsize)
    out, (h, c) = self.lstm(Y, (h0, c0))
    z = h.permute(1, 0, 2).mean(1)
    return z

class LSTMDecoder(nn.Module):
  """
  A general class for stacked lstm decoder
  """
  def __init__(self, output_length, output_size, hidden_size, latent_size, num_layers, directions):
    """
    inputs:
        output_length: length of the output (reconstructed) series
        output_size: dimension of the output (reconstructed) series
        latent_size: dimension of latent representation
        hidden_size: dimension of lstm hidden cells
        num_layers: number of stacked lstm
        directions: 1 for ordinary and 2 for bi-directional lstm
    """
    super().__init__()

    is_bidirectional = (directions == 2)
    
    self.lstm = nn.LSTM(latent_size, hidden_size=hidden_size,
                        bidirectional=is_bidirectional, num_layers=num_layers,
                        proj_size=output_size)

    self.latent_size = latent_size
    self.num_layers = num_layers
    self.directions = directions
    self.output_length = output_length
    self.output_size = output_size
    self.hidden_size = hidden_size
  
  def get_initial(self, batch_size):
    h_size = (self.directions * self.num_layers, batch_size, self.output_size)
    c_size = (self.directions * self.num_layers, batch_size, self.hidden_size)
    h0 = torch.zeros(h_size)
    c0 = torch.zeros(c_size)
    return h0, c0
  
  def forward(self, latent):
    bsize = latent.size(0)
  
    lstm_in = latent.unsqueeze(1).repeat(1, self.output_length, 1)
    lstm_in = lstm_in.permute(1, 0, 2)

    h0, c0 = self.get_initial(bsize)

    out, _ = self.lstm(lstm_in, (h0, c0))
    # average the result of two directions
    if self.directions == 2:
      out = 0.5 * (out[:, :, :self.output_size] + out[:, :, self.output_size:])
    out = out.permute(1, 0, 2)
    return out


class GRUEncoder(nn.Module):
  """
  A general class for stacked GRU encoders
  """
  def __init__(self, input_size, latent_size, hidden_size, num_layers, directions):
    """
    inputs:
        input_size: dimension of input series
        latent_size: dimension of latent representation
        hidden_size: dimension of lstm hidden cells
        num_layers: number of stacked lstm
        directions: 1 for ordinary and 2 for bi-directional lstm
    """
    super().__init__()

    is_bidirectional = (directions == 2)

    self.gru = nn.GRU(input_size, hidden_size=hidden_size,
                      bidirectional=is_bidirectional, num_layers=num_layers)
    
    self.proj = nn.Linear(hidden_size, latent_size)

    self.latent_size = latent_size
    self.num_layers = num_layers
    self.directions = directions
    self.hidden_size = hidden_size
  
  def get_initial(self, batch_size):
    h_size = (self.directions * self.num_layers, batch_size, self.hidden_size)
    h0 = torch.zeros(h_size)
    return h0
  
  def forward(self, Y):
    bsize = Y.size(0)
    Y = Y.permute(1, 0, 2)
    h0 = self.get_initial(bsize)
    _, h = self.gru(Y, h0)
    z = h.permute(1, 0, 2).mean(1)
    z = self.proj(z)
    return z

class GRUDecoder(nn.Module):
  """
  A General class for stacked gru decoder
  """
  def __init__(self, output_length, output_size, hidden_size, latent_size, num_layers, directions):
    """
    inputs:
        output_length: length of the output (reconstructed) series
        output_size: dimension of the output (reconstructed) series
        latent_size: dimension of latent representation
        hidden_size: dimension of gru hidden cells
        num_layers: number of stacked gru
        directions: 1 for ordinary and 2 for bi-directional gru
    """
    super().__init__()

    is_bidirectional = (directions == 2)
    
    self.gru = nn.GRU(latent_size, hidden_size=hidden_size,
                      bidirectional=is_bidirectional, num_layers=num_layers)
    
    self.proj = nn.Linear(hidden_size, output_size)

    self.latent_size = latent_size
    self.num_layers = num_layers
    self.directions = directions
    self.output_length = output_length
    self.output_size = output_size
    self.hidden_size = hidden_size
  
  def get_initial(self, batch_size):
    h_size = (self.directions * self.num_layers, batch_size, self.hidden_size)
    h0 = torch.zeros(h_size)
    return h0
  
  def forward(self, latent):
    bsize = latent.size(0)
  
    gru_in = latent.unsqueeze(1).repeat(1, self.output_length, 1)
    gru_in = gru_in.permute(1, 0, 2)

    h0 = self.get_initial(bsize)

    out, _ = self.gru(gru_in, h0)
    # average the result of two directions
    if self.directions == 2:
      out = 0.5 * (out[:, :, :self.hidden_size] + out[:, :, self.hidden_size:])
    out = out.permute(1, 0, 2)
    out = self.proj(out)
    return out


class MLPEncoder(nn.Module):
  """
  A general class for MLP encoder
  """
  def __init__(self, input_size, input_length, latent_size, hidden_layers=(32,), activation=None):
    """
    inputs:
        input_size: dimension of input series
        input_length : length of input series
        latent_size: dimension of latent representation
        hidden_layers: a tuple of hidden layers dimension
        activation: a custom activation function which can be any PyTorch module supporting a backward pass,
                    if None passed, then the nn.Tanh() will be used
    """
    super().__init__()

    if activation == None:
      activation = nn.Tanh

    # input layer
    mlp_layers = [nn.Linear(input_size*input_length, hidden_layers[0])]
    mlp_layers.append(activation())
   
    # hidden layers
    for layer in range(len(hidden_layers)-1):
      mlp_layers.append(nn.Linear(hidden_layers[layer], hidden_layers[layer+1]))
      mlp_layers.append(activation())
    
    # output layer
    mlp_layers.append(nn.Linear(hidden_layers[-1], latent_size))
    mlp_layers.append(activation())

    self.mlp = nn.Sequential(*mlp_layers)
    
    self.latent_size = latent_size
  
  
  def forward(self, Y):
    Y = Y.flatten(1)
    z = self.mlp(Y)
    return z

class MLPDecoder(nn.Module):
  """
  A general class for MLP encoder
  """
  def __init__(self, input_size, input_length, latent_size, hidden_layers=(32), activation=None):
    """
    inputs:
        input_size: dimension of input series
        input_length : length of input series
        latent_size: dimension of latent representation
        hidden_layers: a tuple of hidden layers dimension
        activation: a custom activation function which can be any PyTorch module supporting a backward pass,
                    if None passed, then the nn.Tanh() wiil be used
    """
    super().__init__()

    if activation == None:
      activation = nn.Tanh
    
    # input layer
    mlp_layers = [nn.Linear(latent_size, hidden_layers[0])]
    mlp_layers.append(activation())
   
    # hidden layers
    for layer in range(len(hidden_layers)-1):
      mlp_layers.append(nn.Linear(hidden_layers[layer], hidden_layers[layer+1]))
      mlp_layers.append(activation())
    
    # output layer
    mlp_layers.append(nn.Linear(hidden_layers[-1], input_size*input_length))
    mlp_layers.append(activation())

    self.mlp = nn.Sequential(*mlp_layers)
    self.unflatten = nn.Unflatten(1, (input_length, input_size))

    self.latent_size = latent_size
  
  
  def forward(self, latent):
    Y = self.mlp(latent)
    Y = self.unflatten(Y)
    return Y


class Encoder_Decoder_TCN(nn.Module):
  """
  A general class for Encoder decoder with 
  dilated Temporal Convolutional Networks (TCN).
  *NOTE*: make sure `input_size` is devisible by $2 ^ {||{hidden\_layers}||_0}$.
  """
  def __init__(self, input_size, input_length, hidden_layers=(128,64),
               kernel_size =8, dilation=2, activation=None, dropout=0.3):
    """
    inputs:
        input_size    : dimension of input series
        input_length  : length of input series
        hidden_layers : a tuple of hidden layers dimension.
                           It's size determines model's depth, e.g., hidden_layers=(128,64,32) has depth three.
        kernel_size   : a size of the convolving kernel
        dilation      : Spacing between kernel elements. Default: 2
        activation    : a custom activation function which can be any PyTorch module supporting a backward pass,
                          if None passed, then the nn.Tanh() will be used
        dropout       : probability of an element to be zeroed. Default: 0.3
    """
    
    if input_length % 2**len(hidden_layers)!=0:
      raise ValueError(f"'Time series length' has to be divisible by number of "\
                       f"'hidden_layers', but {input_length} is not divisible by {2**len(hidden_layers)}!")
    
    super().__init__()
    depth = len(hidden_layers)
    if activation == None:
      activation = nn.Tanh

    
    ##    Encoder: 
    model = []
    for i in range(depth):
      dilation_size = 2 ** i
      in_channels = input_size if i == 0 else hidden_layers[i-1]
      model.append(nn.Conv1d(in_channels, hidden_layers[i],
                             kernel_size, padding='same', dilation=dilation_size))
      model.append(nn.Dropout(dropout))
      model.append(activation())
      model.append(nn.MaxPool1d(2))  
    self.encoder = nn.Sequential(*model)
    self.encoder.latent_size = hidden_layers[-1]

    ##    Decoder:
    model = []
    for i in range(depth-1,-1,-1):
      dilation_size = 2 ** i
      out_channels = input_size if i == 0 else hidden_layers[i-1]
      model.append(nn.Upsample(scale_factor=2))
      model.append(nn.Conv1d(hidden_layers[i], out_channels,
                             kernel_size=kernel_size, padding='same', dilation=dilation_size))
      model.append(nn.Dropout(dropout))
      model.append(activation())

    self.decoder = nn.Sequential(*model)

class AutoEncoder(nn.Module):
  """
  General AutoEncoder class
  """
  def __init__(self, encoder, decoder):
    """
    Args:
      encoder: a PyTorch module
      decoder: a PyTorch module
    """
    super().__init__()

    self.latent_size = encoder.latent_size

    self.encoder = encoder
    self.decoder = decoder

    self.rec_loss = nn.MSELoss()

  def encode(self, Y, inference=False):
    """
    encodes a mini batch of time series
    Args:
      Y : time series batch a PyTorch Tensor (batch_size x seires_length x series_dim)
      inference : if True, only forward pass will happen and the gradient won't be computed
    """
    if inference:
      self.encoder.eval()
      with torch.no_grad():
        return self.encoder(Y)
    else:
      return self.encoder(Y)

  def decode(self, latent):
    """
    decodes a mini batch of latent vectors
    Args:
      latent: a PyTorch Tensor (batch_size x latent_dim)
    """
    return self.decoder(latent)

  def loss(self, minibatch):
    latent = self.encode(minibatch['Y'], inference=False)
    reconstructed = self.decode(latent)
    loss = self.rec_loss(minibatch['Y'], reconstructed)
    return loss


class PyTorchTrainer():
  """
  Abstract trainer for PyTorch models
  """
  def __init__(self, model, batch_size=16, learning_rate=0.02):
    """
    Args:
      model : a PyTorch module
      batch_size: size of each mini batch
      learning_rate: optimizer's learning rate
    """
    self.model = model
    self.batch_size = 16
    self.learning_rate = learning_rate
    self.loss_callbacks = []

    self.initialize()
  
  def initial_optimizer(self):
    """
    initializing the optimizer
    """
    return torch.optim.Adam(self.model.parameters(), lr=self.learning_rate)
  
  def initialize(self):
    self.optimizer = self.initial_optimizer()

  def register_loss_callback(self, fn):
    """
    register a callback to which the loss will be passed at each iteration
    """
    if callable(fn):
      self.loss_callbacks.append(fn)
    else:
      raise ValueError("The input is not a valid callable object")

  def apply_loss_callbacks(self, loss):
    """
    calling all registered callbacks
    """
    for fn in self.loss_callbacks:
      fn(loss)
  
  def get_mini_batch(self):
    """
    a method which must be implemented to provide a mini batch for model
    it's better to use a dictionary as mini batch
    """
    raise NotImplementedError("You must provide a method for sampling a batch")
  

  def step(self):
    """
    A single training step
    """
    # train mode
    self.model.train()
    # set thte accumulated gradient to zero 
    self.optimizer.zero_grad()
    # computing the loss for a mini batch
    mini_batch = self.get_mini_batch()
    loss = self.model.loss(mini_batch)
    # backprop and a single optimization step
    loss.backward()
    self.optimizer.step()
    # calling the registeres loss callbacks
    self.apply_loss_callbacks(loss.item())



import numpy as np
import pandas as pd
import tsfresh
from metats.features import FeatureGenerator


class TsFresh(FeatureGenerator):
    """
    Generate features using TsFresh
    """
    def fit(self, Y):
        pass
    
    def reshape_for_tsfresh(self, Y):
        """
        Reshape the timeseries for tsfresh
        inputs:
            Y: the timeseries (numpy array) (num_series x series_length x covariates_dim)
        """
        reshaped = []
        for id in range(Y.shape[0]):
            for time in range(Y.shape[1]):
                row = np.array([id, time])
                row = np.hstack([row, Y[id, time, :]])
                reshaped.append(row)
        reshaped = np.vstack(reshaped)
        return reshaped

    def transform(self, Y):
        """
        Extract features using tsfresh
        inputs:
            Y: the timeseries (numpy array) (num_series x series_length x covariates_dim)
        """
        reshaped = self.reshape_for_tsfresh(Y)
        covariates_dim = Y.shape[2]
        column_names = ['id', 'time'] + ['v_'+ str(i) for i in range(covariates_dim)]
        tsf = pd.DataFrame(reshaped, columns=column_names)
        del reshaped
        features = tsfresh.extract_features(tsf, column_id='id', column_sort='time')
        # remove the Nan values
        features.dropna(axis=1, inplace=True)
        features = features.values[:, 2:]
        return features




import numpy as np

class MetaLearning():
    """
    Base Class for Meta-Learning Pipelines
    """
    def __init__(self, method='averaging', loss='mse', reduction='none', reduction_dim=64):
        """
        Args:
            method: can be either 'averaging' or 'selection'. in the case of averagin,
                    the predictions of base-forecaster will be combined by weighting
                    based on classification weights. In the 'selection' mode, the best
                    forecaster will be selected and others will be ingored
            loss: loss function for generating meta-learning labels, can be 'mse', 'mae'.
                  to use a custom loss function you can also pass a callable object.
            reduction: you can apply dimensionality reduction on generated features.
                       can be 'none', 'pca'. use reduction dim to specify
                       final dimension.
            reduction_dim: will be ignored if reduction='none'.
                   
        """
        self.method = method
        self.feature_generators = []
        self.base_forecasters = []
        self.base_forecasters_name = []
        self.meta_learner = None
        self.loss = loss
        self.reduction = reduction
        self.reduction_dim = reduction_dim

        if self.reduction == 'pca':
            from sklearn.decomposition import PCA
            self.reducer = PCA(n_components=self.reduction_dim)
    
    def add_feature(self, feature_generator):
        """
        Add a feature generator to the pipeline
        """
        self.feature_generators.append(feature_generator)
    
    def add_forecaster(self, forecaster, forecaster_name=""):
        """
        Add a base forecaster to the pipeline
        """
        from metats.forecasters.utils import is_sktime_forecaster, is_darts_forecaster, is_nixtla_stats_forecaster
        # checking for dart forecasters
        wrapped = forecaster
        if is_darts_forecaster(forecaster):
            from metats.forecasters.wrappers import DartsForecasterWrapper
            wrapped = DartsForecasterWrapper(forecaster)
        elif is_sktime_forecaster(forecaster):
            from metats.forecasters.wrappers import SKTimeForecasterWrapper
            wrapped = SKTimeForecasterWrapper(forecaster)
        elif is_nixtla_stats_forecaster(forecaster):
            from metats.forecasters.wrappers import NixtlaStatsForecasterWrapper
            wrapped = NixtlaStatsForecasterWrapper(forecaster)

        # append the wrapped forecaster to the stack
        self.base_forecasters.append(wrapped)
        # generate a name for forecater
        name = "Forecaster {}".format(len(self.base_forecasters))
        if len(forecaster_name) > 0:
            name = forecaster_name
        self.base_forecasters_name.append(name)

        

    
    def loss_fn(self, y_true, y_pred):
        """
        Loss function for the meta-learner
        Args:
            y_true: true values (numpy array) (num_series x horizon)
            Y_pred: the matrix of predictions (num_series x num_models x horizon)
        """
        if self.loss == 'mse':
            return np.mean(np.square(y_true - y_pred), axis=2)
        elif self.loss == 'mae':
            return np.mean(np.abs(y_true - y_pred), axis=2)
        elif callable(self.loss):
            return self.loss(y_true, y_pred)
        else:
            raise ValueError('Loss function not supported')

    def labels_selection(self, y_true, Y_pred, return_one_hot=False):
        """
        Generate labels for the meta-learner using selection
        Args:
            y_true: true values (numpy array) (num_series x horizon)
            Y_pred: the matrix of predictions (num_series x num_models x horizon)
        """
        error = self.loss_fn(y_true, np.transpose(Y_pred, (1, 0, 2)))
        error = error.T
        label_indices = np.argmin(error, axis=1)
        if return_one_hot:    
            # convert to one-hot encoding
            labels = np.zeros(Y_pred.shape[1:])
            for i in range(Y_pred.shape[1]):
                labels[i, label_indices[i]] = 1.0
            return labels
        else:
            return label_indices

    def generate_labels(self, y_true, Y_pred):
        """
        Generate labels for the meta-learner
        Args:
            y_true: true values
            Y_pred: the matrix of predictions 
        """
        if self.method == 'selection':
            return self.labels_selection(y_true, Y_pred)
        elif self.method == 'averaging':
            return self.labels_selection(y_true, Y_pred)
        else:
            raise ValueError('Method not supported')
        
    def generate_features(self, Y, prediction=False):
        # generate meta features
        meta_features = []
        for feature_generator in self.feature_generators:
            if (not prediction) or (not feature_generator.is_trainable()):
                feature_generator.fit(Y) 
            meta_features.append(feature_generator.transform(Y))
        return np.hstack(meta_features)
    
    def generate_prediction(self, forecaster, Y, fh, forecast_dim):
        """
        Generate predictions for a single forecaster
        Args:
            forecaster: the forecaster to be used
            Y: the timeseries (numpy array) (num_series x series_length x covariates_dim)
            fh: forecast horizon
            forecast_dim: the dimension of the variable to be forecasted
        """
        return forecaster.predict(Y, fh, forecast_dim)
    
    def generate_predictions(self, Y, fh, forecast_dim=0):
        # generate predictions
        predictions = []
        for base_forecaster in self.base_forecasters:
            prediction = self.generate_prediction(base_forecaster, Y, fh, forecast_dim) 
            if prediction.ndim == 3:
                prediction = prediction[:, :, forecast_dim]
            prediction = np.expand_dims(prediction, axis=1)
            predictions.append(prediction)
        predictions = np.concatenate(predictions, axis=1)
        return predictions
    
    def add_metalearner(self, metalearner):
        """
        Add a meta-learner to the pipeline
        Args:
            metalearner: the meta-learner to be added a scikit-learn estimator
        """
        self.meta_learner = metalearner

    def reduce_meta_features(self, meta_features, prediction=False):
        """
        applying dimensionality reduction on the generated features 
        Args:
            meta_features: numpy array of meta-features (num_series, features_dim)
            prediction: if True the stored features from training stage will be used
        """
        num_training_series = 0
        if prediction:
            num_training_series = self.training_features.shape[0] 
            # concatenating the training features
            meta_features = np.vstack([self.training_features, meta_features])
        reduced = self.reducer.fit_transform(meta_features)[num_training_series:, :]
        return reduced

    def fit(self, Y, fh, forecast_dim=0):
        """
        Fit the meta-learner
        Args:
            Y: the timeseries (numpy array) (num_series x series_length x covariates_dim)
            fh: forecasting horizon
            forecast_dim: forecasting dimension
        """
        # data spliting for meta-learning
        Y_true = Y[:, -fh:, forecast_dim]

        self.features_fh = fh
        meta_features = self.generate_features(Y[:, :-fh, :], prediction=False)
        predictions = self.generate_predictions(Y[:, :-fh, :], fh, forecast_dim=forecast_dim)
        labels = self.generate_labels(Y_true, predictions)
        
        # saving the generated features for using in prediction phase
        if self.reduction != 'none':
            self.training_features = np.copy(meta_features)
            meta_features = self.reduce_meta_features(meta_features)

        self.meta_learner.fit(X=meta_features, y=labels)

    def predict_generate_weights(self, meta_features):
        """
        Predict using the meta-learner
        Args:
            metafeatures: the extracted meta-features (numpy array) (num_series x features_dim)
        """
        weights = None
        if self.method == 'selection':
            weights = self.meta_learner.predict(meta_features)
        elif self.method == 'averaging':
            weights = self.meta_learner.predict_proba(meta_features)
        
        return weights

    def averaging_predictions(self, weights, predictions):
        """
        Generating the predictions by averaging each base-forecaster
        """
        weighted_predictions = []
        for series in range(predictions.shape[0]):
            p = weights[series, :].reshape(1, -1) @ predictions[series, :, :]
            weighted_predictions.append(p)
        result = np.vstack(weighted_predictions)
        return result
    
    def selection_predictions(self, weights, predictions):
        """
        Generating the predictions by selecting the best-forecaster
        """
        selected_predictions = []
        for series in range(predictions.shape[0]):
            p = predictions[series, weights[series], :].reshape(1,-1)
            selected_predictions.append(p)
        result = np.vstack(selected_predictions)
        return result

    def predict(self, Y, fh, forecast_dim=0, return_metalearning_data=False):
        """
        Predict using the meta-learner
        Args:
            Y: the timeseries (numpy array) (num_series x series_length x covariates_dim)
            fh: forecast horizon for predicting
            forecast_dim: the dimension of the variable to be forecasted
            return_metalearning_data: if True, the meta-features, weights and base forecasts will be returned
        Return:
            result: (numpy array) (num_series x forecast_horizon)
            if return_metalearning_data is True:
                predictions: (numpy array) (num_series x number_of_base_forecasters x forecast_horizon)
                weights: (numpy array) (num_series x number_of_base_forecasters)
                meta_features: (numpy array) (num_series x features_dim)
        """
        meta_features = self.generate_features(Y[:, :-self.features_fh, :], prediction=True)

        if self.reduction != 'none':
            meta_features = self.reduce_meta_features(meta_features, prediction=True)

        predictions = self.generate_predictions(Y, fh, forecast_dim=forecast_dim)
        weights = self.predict_generate_weights(meta_features)

        if self.method == 'averaging':
            result = self.averaging_predictions(weights, predictions)
        else:
            result = self.selection_predictions(weights, predictions)

        if return_metalearning_data:
            return result, predictions, weights, meta_features
        
        return result




import numpy as np
import torch
import torch.nn as nn
from metats.features import FeatureGenerator
from metats.features.deep import PyTorchTrainer


class DeepAETrainer(PyTorchTrainer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_mini_batch(self):
        data_size = self.Y.shape[0]
        sample_size = min(self.batch_size, data_size)
        indices = np.random.choice(data_size, sample_size, replace=False)
        mini_batch = self.Y[indices]
        mini_batch_tensor = torch.from_numpy(mini_batch).float()
        return {'Y' : mini_batch_tensor}

class DeepAutoEncoder(FeatureGenerator):
    """
    A wrapper class for feature extraction using deep auto encoders
    """
    def __init__(self, auto_encoder, verbose=False, epochs=64, learning_rate=0.006, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.auto_encoder = auto_encoder
        self.epochs = epochs
        self.verbose = verbose
        self.learning_rate = learning_rate
    
    def print_loss_callback(self, loss):
        print('Deep AE Reconstruction Loss: {:.3f}'.format(loss))


    def fit(self, Y):
        trainer = DeepAETrainer(model=self.auto_encoder, learning_rate=self.learning_rate)
        if self.verbose:
            trainer.register_loss_callback(self.print_loss_callback)

        trainer.Y = Y
        for epoch in range(self.epochs):
            trainer.step()
    
    def transform(self, Y):
        Y_tensor = torch.from_numpy(Y).float()
        latent = self.auto_encoder.encode(Y_tensor, inference=True)
        return latent.numpy()
    
    def is_trainable(self):
        return True

class Contrastive(FeatureGenerator):
    """
    Base Class for Contrastive feature extractors
    """
    def fit(self, Y):
        pass
    
    def transform(self, Y):
        pass


class FeatureGenerator():
    """
    Base class for feature generators
    """
    def fit(self, timeseries):
        """
        Fit the feature generator
        inputs:
            timeseries: the timeseries (numpy array) (num_series x series_length)
        """
        raise NotImplementedError('FeatureGenerator.fit() not implemented')
    
    def transform(self, timeseries):
        """
        Generate features
        inputs:
            timeseries: the timeseries (numpy array) (num_series x series_length)
        """
        raise NotImplementedError('FeatureGenerator.transform() not implemented')

    def fit_transform(self, timeseries):
        self.fit(timeseries)
        return self.transform(timeseries)

    def is_trainable(self):
        return False


import pytest
import numpy as np
from metats.features.statistical import TsFresh

def test_tsfrsh_transform():
    data = np.random.rand(10, 55, 5)
    try:
        tsfresh = TsFresh()
    except:
        pytest.fail("Failed to instantiate TsFresh")
    
    tsfresh.fit(data)
    try:
        features = tsfresh.transform(data)
    except:
        pytest.fail("TSFresh.transform() failed")

    assert features.shape[0] == 10



import numpy as np

class Dataset():
    """
    Base class for datasets
    """
    pass


class ETSDataset(Dataset):
    """
    ETS dataset
    Sampling synthetic data from the ETS model
    """
    def __init__(self, ets_families, length=100, freq=7, positive=True):
        """
        Args:
            ets_families: list of ETS families to sample from and number of samples
            length: length of generated series
            freq : frequency
            positive : if True, only positive-valued series will be generated 
        """
        super().__init__()
        
        self.ets_families = ets_families
        # generating series
        self.series = []
        for family, n_series in ets_families.items():
            for i in range(n_series):
                error, trend, seasonality = family.replace(' ', '').split(',')
                if trend == 'Ad':
                    phi = np.random.uniform(low=0.95, high=1.0)
                else:
                    phi = 1.0

                series = self.generate_series(length=length, alpha=np.random.uniform(),
                                beta=np.random.uniform(),
                                phi=phi,
                                gamma=np.random.uniform(),
                                l0 = np.random.normal(),
                                b0 = np.random.normal(),
                                seasonality_freq=freq,
                                s_init=np.random.normal(size=(freq)),
                                error=error, trend=trend, seasonality=seasonality)
                
                if positive:
                    series += max(0, -series.min())
                self.series.append(series)

    def load(self, return_family=False):
        """
        Load dataset
        """
        series = np.array(self.series)
        if return_family:
            families = []
            for family, n_series in self.ets_families.items():
                for i in range(n_series):
                    families.append(family)
            return series, families
        return series

    def generate_series(self, error='A', trend='A', seasonality='A', length=100, alpha=0.6, beta=0.6, gamma=0.5, phi=1.0,
               l0=0.2, b0=0.1, error_mu=0.0, error_sigma=0.01, seasonality_freq=7, s_init=np.random.normal(size=7)):
        """
        samples a time series from ETS class
        input parameters:
        - length: length of the time series
        - error : type of the error compoenent
                'A' for additive error
                'M' for multiplicative error
        - trend : type of the trend component
                'N' for no trend
                'A' for additive trend
                'Ad' for additive damped trend
        - seasonality: type of the seasonality component
                'N' for no seasonality
                'A' for additive seasonality
                'M' for multiplicativeseasonality
        - seasonality_freq : frequency of the seasonality
        - alpha, beta, gamma, phi, l0, b0, error_mu, error_sigma : parameters of the ETS model
        """
        series = []
    
        y = []
        l = [l0]
        b = [b0]
        s = list(s_init)
        
        m = seasonality_freq
        
        if trend == 'N':
            phi = 0
        
        for i in range(length):
            et = np.random.normal(error_mu, error_sigma)
            if seasonality == 'N':
                sm = 0
            else:
                sm = s[-m]


            # additive seasonality
            if seasonality=='A' or seasonality=='N':
                st = sm
                bt = phi*b[-1]
                lt = l[-1] + phi*b[-1]
                yt = l[-1] + phi*b[-1] + sm

                # additive error
                if error == 'A':
                    st += gamma*et
                    lt += alpha*et
                    bt += beta*et
                    yt += et
                # multiplicative error
                else:
                    e_temp = l[-1] + b[-1] + sm
                    st += gamma*e_temp*et
                    bt += beta*e_temp*et
                    lt += alpha*e_temp*et
                    yt *= (1+et)

            # multiplicative seasonality
            elif seasonality=='M':
                st = sm
                bt = phi*b[-1]
                lt = l[-1] + phi*b[-1]
                yt = (l[-1] + phi*b[-1])*sm

                # additive error
                if error == 'A':
                    st += gamma*et/(l[-1] + phi*b[-1])
                    lt += alpha*et/sm
                    bt += beta*et/sm
                    yt += et
                # multiplicative error
                else:
                    st *= 1 + gamma*et
                    bt += beta*(l[-1] + phi*b[-1])*et
                    lt *= 1 + alpha*et
                    yt *= 1 + et

            y.append(yt)
            l.append(lt)
            b.append(bt)
            s.append(st)
        
        return np.array(y)


import pytest
import numpy as np
import metats
from sktime.forecasting.naive import NaiveForecaster
# from metats.features.statistical import TsFresh
from metats.pipeline import MetaLearning
from metats.features import FeatureGenerator
from sklearn.ensemble import RandomForestClassifier
from sktime.forecasting.compose import make_reduction
from sklearn.neighbors import KNeighborsRegressor

class CustomFeature(FeatureGenerator):
    def fit(self, Y):
        pass
    
    def transform(self, Y):
        return np.random.normal(size=(Y.shape[0], 3))

def custom_loss(y_true, y_pred):
    return np.mean(np.square(y_true - y_pred), axis=2)


def test_basic_pipeline():
    data = np.random.normal(size=(2, 55, 5))
    regressor = KNeighborsRegressor(n_neighbors=1)
    forecaster1 = make_reduction(regressor, window_length=15, strategy="recursive")
    forecaster2 = NaiveForecaster()    


    feature1 = CustomFeature()

    pipeline = MetaLearning(method='selection', loss=custom_loss)
    pipeline.add_forcecaster(forecaster1)
    pipeline.add_forcecaster(forecaster2)
    assert len(pipeline.base_forecasters) == 2
    pipeline.add_feature(feature1)
    assert len(pipeline.feature_generators) == 1
    pipeline.add_metalearner(RandomForestClassifier())
    assert isinstance(pipeline.meta_learner, RandomForestClassifier)
    pipeline.fit(data, fh=7)
    weights, predictions = pipeline.predict(data, fh=7)
    print(weights.shape)
    print(weights)
    print(predictions.shape)
    print(predictions)





import numpy as np
import pandas as pd
from metats.forecasters import BaseForecaster
from metats.forecasters.utils import is_sktime_forecaster, is_darts_forecaster, is_nixtla_stats_forecaster



class DartsForecasterWrapper(BaseForecaster):
    """
    A Wrapper class for darts models
    """
    def __init__(self, darts_model):
        """
        inputs:
            dart_model: a darts models which must be checked using is_darts_forecaster
        """
        if not is_darts_forecaster(darts_model):
            raise ValueError('The input is not a valid darts forecaster')

        self.darts_model = darts_model
    
    def predict(self, Y, fh, forecast_dim):
        """
        inputs:
            Y: the timeseries (numpy array) (num_series x series_length x covariates_dim)
            fh: forecast horizon
            forecast_dim: the dimension of the forecast to be generated
        """
        from darts import TimeSeries as DTS
        
        predictions = []
        # forecast for each series
        for i in range(Y.shape[0]):
            frame = DTS.from_values(Y[i, :, forecast_dim:forecast_dim+1])
            self.darts_model.fit(frame)
            single_pred = np.squeeze(self.darts_model.predict(fh))
            predictions.append(single_pred.reshape((1, -1)))
        predictions = np.concatenate(predictions, axis=0)
        return predictions

class SKTimeForecasterWrapper(BaseForecaster):
    """
    A Wrapper class for sktime models
    """
    def __init__(self, sktime_model):
        """
        inputs:
            dart_model: a sktime models which must be checked using is_sktime_forecaster
        """
        if not is_sktime_forecaster(sktime_model):
            raise ValueError('The input is not a valid sktime forecaster')
        self.sktime_model = sktime_model
    
    def predict(self, Y, fh, forecast_dim):
        """
        inputs:
            Y: the timeseries (numpy array) (num_series x series_length x covariates_dim)
            fh: forecast horizon
            forecast_dim: the dimension of the forecast to be generated
        """
        if isinstance(fh, int):
            fh = 1 + np.arange(fh)
        
        predictions = np.zeros((Y.shape[0], len(fh)))
        # forecast for each series
        for i in range(Y.shape[0]):
            self.sktime_model.fit(y=Y[i, :, forecast_dim], fh=fh)
            predictions[i, :] = np.squeeze(self.sktime_model.predict())
        
        return predictions

class NixtlaStatsForecasterWrapper(BaseForecaster):
    """
    A Wrapper class for Nixtla statistical forecasters
    """
    def __init__(self, nixtla_model):
        """
        inputs:
            nixtla_model: a darts models which must be checked using is_nixtla_stats_forecaster
        """
        if not is_nixtla_stats_forecaster(nixtla_model):
            raise ValueError('The input is not a valid nixtla forecaster')

        self.nixtla_model = nixtla_model
    
    def predict(self, Y, fh, forecast_dim):
        """
        inputs:
            Y: the timeseries (numpy array) (num_series x series_length x covariates_dim)
            fh: forecast horizon
            forecast_dim: the dimension of the series to be generated
        """

        # prepare data for nixtla
        data_df = pd.DataFrame(Y[:, :, forecast_dim])
        data_df['unique_id'] = data_df.index
        nixtla_df = data_df.melt(var_name='ds', value_name='y', id_vars=['unique_id'])
        # fit the nixtla frecaster
        self.nixtla_model.fit(nixtla_df)
        # generting and collecting the predictions
        pred_df = self.nixtla_model.predict(h=fh)
        predictions = []
        for i in range(Y.shape[0]):
            single_pred = pred_df[pred_df.index==i][pred_df.columns[-1]].values
            predictions.append(single_pred.reshape((1, -1)))
        predictions = np.concatenate(predictions, axis=0)
        return predictions